#include <iostream>
#include "utMath.h"
#include <btBulletDynamicsCommon.h>
#include <cstdio>

#include "Game.h"
#include "Bullet.h"
#include "Physics.h"

/**
 * constructs a bullet with a graphics model and the maximum flying distance. In order to disconnect the
 * bullet from the camera (which would disturb the bullet flight vector) a proxy node is used which must
 * be given from the calling class.
 * @param parentNodeProxy a proxy is a group node and acts as a fixed point for shot bullets.
 *		this proxy must contain the player's position and rotation at shooting time.
 *		this H3DNode is removed inside this class so the shooting class doesn't need to hold an
 *		instance of this proxy model.
 * @param bulletRes this is the ressource to the bullet model
 * @param distance this is the maximum distance that this bullet may fly. please note,
 *		that additionally the bullet deactivates after a certain period of time.
 */
Bullet::Bullet(Physics* physics, H3DNode parentNodeProxy, H3DRes bulletRes, const float distance) {
	_activeTime = 0;
	_distanceToFly = distance;
	MAX_FLY_TIME_S = 2.5f;

	//create graphics bullet
	_model = h3dAddNodes(parentNodeProxy, bulletRes);
	_bulletProxy = parentNodeProxy;
	float ntx, nty, ntz;
	float nrx, nry, nrz;
	float nsx, nsy, nsz;
	h3dGetNodeTransform(_model, &ntx, &nty, &ntz, &nrx, &nry, &nrz, &nsx, &nsy, &nsz);

	h3dSetNodeTransform(_model,
		ntx + 1, nty, ntz,
		-90, 10, nrz, //rotation
		0.05f, 0.05f, 0.05f); //size
}

Bullet::~Bullet() {
}

/**
 * updates the graphics state of this bullet, that is: let the bullet fly 
 * in shooting direction
 * @param timeStep the delta time to be rendered
 */
void Bullet::update(const float timeStep) {
	_activeTime += timeStep;
	
	float speed = ::Values::BULLET_SPEED_M_S * _activeTime;
//std::printf("from: %s\nto:  %s\n", _debug.toString().c_str(), _start.toString().c_str());

	float ntx, nty, ntz;
	float nrx, nry, nrz;
	float nsx, nsy, nsz;
	h3dGetNodeTransform(_bulletProxy, &ntx, &nty, &ntz, &nrx, &nry, &nrz, &nsx, &nsy, &nsz);

	// Set camera parameters
	h3dSetNodeTransform(_model, 
		0, 0, -speed, 
		-90, 10, nrz, 
		nsx, nsy, nsz);

//std::printf("fr: %s\nto:  %s\n", prx.toString().c_str(), bul.toString().c_str());	
	//TODO maybe leave a particle system trace
}

/**
 * returns whether this bullet is still flying or not
 * @return true if this bullet is not yet deactivated by either
 *		time or distance
 */
bool Bullet::isActive() {
//std::printf("current time %g\n", _activeTime);
	if(_activeTime < MAX_FLY_TIME_S && _distanceToFly > 0) {
		return true;
	}
	
	h3dRemoveNode(_model);
	h3dRemoveNode(_bulletProxy);
	return false;
}

//reserved for future usage
///**
// * sets the current player position in world space
// */
//void Bullet::setPosition(const Vec3& newPosition) {
//	btTransform transform;
//	transform = _body->getCenterOfMassTransform();
//	transform.setOrigin(btVector3(newPosition.x, newPosition.y, newPosition.z));
//	_body->setCenterOfMassTransform(transform);
//}