#include "Horde3D.h"
#include "Horde3DUtils.h"
#include <btBulletDynamicsCommon.h>

#include "Physics.h"
#include "Game.h"
#include "World.h"
#include "Explosive.h"
#include "Vec3.h"
#include "CharacterCollisionCallback.h"
#include "IShootable.h"

/**
 * This class provides both physics and graphics simulation of an explosive body, like a gasoline barrel.
 * An Explosive is designed to 
 * @param modelRes
 * @param world
 */
Explosive::Explosive(Game* game, H3DNode parent, Physics* physics, const int id, H3DRes modelRes, const Vec3 position, const Vec3 rotation, const float scale)
		: IShootable(game, parent, physics, id, "barrel") {
	_doExplode = false;
	_health = 3;
	_scale = scale;

	// Create collions shape
	_cylinderRadius = 0.30f;
	float bodyHeight = 1.0f;
	_cylinderHeight = bodyHeight - _levelBelowCylinder;
	btVector3 halfExtent(_cylinderRadius, bodyHeight / 2, _cylinderRadius);
	_cylinder = new btCylinderShape(halfExtent); 

	// Create physics body
	_groundMotionState = new btDefaultMotionState(btTransform(
			btQuaternion(0,0,0,1),
			btVector3(position.x, position.y, position.z)) // Start position of the enemy
	);

	// Create barrel with 50 kg
	_body = new btRigidBody(150, _groundMotionState, _cylinder);
	_body->setFriction(2.0f);
	_body->setDamping(0.5f, 0.5f);
	//some bodies get quickly deactivated... avoid this here
	_body->setActivationState(DISABLE_DEACTIVATION);
	_physics->addShootable(this); //deconst-delete is managed by physics

	H3DRes particleSysRes = h3dAddResource(H3DResTypes::SceneGraph, "particles/explosiveExplosion/explosiveExplosion.scene.xml", 0);
	bool loaded = h3dutLoadResourcesFromDisk(::Values::CONTENT_DIR);
	if(! loaded) std::cout << "could not load explosive ressources" << std::endl;

	//attach explosion to the world
	_emitter = h3dAddNodes(parent, particleSysRes);
	h3dSetNodeTransform(_emitter, position.x, position.y, position.z, 
		90, 0, 0, 
		1, 1, 1);
	h3dSetNodeActivation(_emitter, false);

	//attach model before explosion to the world
	_model = h3dAddNodes(parent, modelRes);
	h3dSetNodeTransform(_model, position.x, position.y, position.z, 
		rotation.x, rotation.y, rotation.z, 
		scale, scale, scale);
}

Explosive::~Explosive() {

}

/**
 * updates the graphics and physics state for explosive entities
 * @param timeStep
 */
void Explosive::update(const float timeStep) {
	//draw only explosion
	if(_doExplode) {
		//don't do anything after this shootable is dead
		//if(h3dHasEmitterFinished(_emitter)) return;

		// Animate particle systems (several emitters in a group node)
		Game::advanceEmitter(_emitter, timeStep);
	}
	else {
		//draw usual graphics in respect to physics simulation
		Vec3 currentPosition = getPosition();
		Vec3 rollbackPosition;
		Vec3 collisionPoint;

		//bool onground = castRayAndCorrectPosition(currentPosition, collisionPoint);
		//if(onground) {
		//	rollbackPosition = currentPosition;
		//	setPosition(currentPosition);
		//	//correct body collision detection to avoid to run thru things
		//	//check if other bodies collide
		//	btConvexShape *shape = (btConvexShape*)_body->getCollisionShape();
		//	btTransform from = _body->getCenterOfMassTransform();
		//	btTransform to = from;
		//	to.setOrigin(to.getOrigin());
		//	// Check collisions
		//	CharacterCollisionCallback cb(_body, from.getOrigin(), to.getOrigin());
		//	_physics->getWorld()->convexSweepTest(shape, from, to, cb);
		//	if (! cb.hasHit()) {
		//		setPosition(currentPosition);

		//		// Do a second raycast to ensure that we still after moving are at the right
		//		// distance to the terrain
		//		// Cast a ray from the bottom of the capsule 1m down, again
		//		castRayAndCorrectPosition(currentPosition, collisionPoint);
		//		setPosition(currentPosition);
		//	}
		//	else {
		//		// currentPosition does not reflect the actual position, reset it.
		//		currentPosition = rollbackPosition;
		//	}
		//}

		btTransform trans;
		_body->getMotionState()->getWorldTransform(trans);
		float explosiveOffset = ((btCapsuleShape*)_body->getCollisionShape())->getHalfHeight();
		//move model
		h3dSetNodeTransform(_model, 
			trans.getOrigin().x(), trans.getOrigin().y(), trans.getOrigin().z(), 
			0 ,0, 0, 
			_scale, _scale, _scale);
		//move explosion-to-be
		h3dSetNodeTransform(_emitter, 
			trans.getOrigin().x(), trans.getOrigin().y(), trans.getOrigin().z(), 
			90, 0, 0, 
			1, 1, 1);
	}
}

/**
 *
 *
 */
void Explosive::setPosition(Vec3 newPosition) {
	btTransform transform;
	transform = _body->getCenterOfMassTransform();
	transform.setOrigin(btVector3(newPosition.x, newPosition.y, newPosition.z));
	_body->setCenterOfMassTransform(transform);
}

/**
 *
 *
 */
Vec3 Explosive::getPosition() {
	btTransform transform;
	transform = _body->getCenterOfMassTransform();
	btVector3 pos = transform.getOrigin();
	return Vec3(pos.x(), pos.y(), pos.z());
}

/**
 *
 * @see IShootable
 */
void Explosive::hit() {
	_health--;
	if(_health <= 0) {
		_doExplode = true;

		//exchange model for explosion
		h3dSetNodeActivation(_model, false);
		h3dSetNodeActivation(_emitter, true);

		//move physics body away and deactivate 
		_body->setActivationState(WANTS_DEACTIVATION);
		setPosition(Vec3(-100,-100,-100));
	}
}

/**
 * Corrects the vertical position of the physics body in respect to the terrain.
 * @param calculatedCurrentPosition the position on the world
 * @param collisionPoint the actual point where the explosive body should cross the world
 * @param physics needed for physics calculation
 */
bool Explosive::castRayAndCorrectPosition(Vec3& calculatedCurrentPosition, Vec3& collisionPoint) {
	float cylinderBottom = (-_cylinderHeight / 2) - _cylinderRadius;

	// Cast a ray from the bottom of the capsule 4m down, again
	Vec3 from = calculatedCurrentPosition;
	Vec3 to = calculatedCurrentPosition;
	from.y += cylinderBottom + 0.5f;
	to.y += cylinderBottom - 0.5f;

	bool onground = _physics->castRay(from, to, collisionPoint, _body);

	// Reset the coordinate so that the capsule is 0.2 m above the collision point
	if (onground) {
		calculatedCurrentPosition = Vec3(
			calculatedCurrentPosition.x, 
			(_cylinderHeight / 2)+ collisionPoint.y,
			calculatedCurrentPosition.z);
	}
	return onground;
}