//includes
#if defined(_MSC_VER) || defined(_WINDOWS_) || defined(_WIN32)
	#include "glfw.h"
#else
	#include <GL/glfw.h>
#endif
#include <vector>
#include <math.h>
#include <iostream>
#include <sstream>
#include <string>
#include "Horde3D.h"
#include "Horde3DUtils.h"
#include "utMath.h"

#include "Game.h"
#include "Physics.h"
#include "Player.h"
#include "Enemy.h"
#include "World.h"

const char*        ::Values::CONTENT_DIR         = "../Content";
const float        ::Values::WALKING_SPEED_KM_H  = 15.0;
const float        ::Values::AIRPLANE_SPEED_KM_H = 100.0f;
const float        ::Values::BULLET_SPEED_M_S    = 500.0f;
const float        ::Values::SHOOTING_DISTANCE   = 100.0f;
const int          ::Values::MAX_HEALTH          = 100;
const int          ::Values::WORLD_SCALE         = 50;
const int          ::Values::MAX_BODY_COUNT      = 10;
const int          ::Values::NUMBER_OF_ENEMIES   = 5;
const float        ::Values::CHAINGUN_RECOIL     = 1.0f;
const float        ::Values::BOLTER_RECOIL       = 0.5f;

/**
 * does currently nothing
 */
Game::Game() {
	_debugViewMode = false;
	_wireframeMode = false;
}

/**
 * keeps the memory clean
 */
Game::~Game() {
	delete _physics;
	delete _player;
	delete _world;

	for(int i = 0; i < _numberOfEnemies; i++) {
		delete _enemies[i];
	}
}

/**
 * inits the graphics engine and creates physics, player, and enemies
 */
bool Game::init() {	
	// Initialize engine
	if(! h3dInit()) {	
		h3dutDumpMessages();
	}
		
	// Initialize physics
	_physics = new Physics();

	// Set options
	h3dSetOption(H3DOptions::LoadTextures, 1);
	h3dSetOption(H3DOptions::TexCompression, 1);
	h3dSetOption(H3DOptions::FastAnimation, 1);
	h3dSetOption(H3DOptions::MaxAnisotropy, 16);
	h3dSetOption(H3DOptions::ShadowMapSize, 8192);
	h3dSetOption(H3DOptions::MaxLogLevel, 4);

	// Add resources
	// Overlays
	_fontMatRes = h3dAddResource(H3DResTypes::Material, "overlays/font.material.xml", 0);
	_panelMatRes = h3dAddResource(H3DResTypes::Material, "overlays/panel.material.xml", 0);
	_crossHairMatRes = h3dAddResource(H3DResTypes::Material, "overlays/crosshair.material.xml", 0);
	// Pipelines
	_forwardPipeRes = h3dAddResource(H3DResTypes::Pipeline, "pipelines/forward.pipeline.xml", 0);
	_reflectionPipeRes = h3dAddResource(H3DResTypes::Pipeline, "pipelines/reflection.pipeline.xml", 0);
	//sky
	H3DRes skyBoxRes = h3dAddResource(H3DResTypes::SceneGraph, "models/skybox/skybox.scene.xml", 0 );
	_waterplaneTexRes = h3dCreateTexture("NewRenderTexture", 512, 512, H3DFormats::TEX_RGBA16F, H3DResFlags::TexRenderable);
	// Load resources
	bool loaded = h3dutLoadResourcesFromDisk(::Values::CONTENT_DIR);

	// Add scene nodes
	// Add camera
	_cam = h3dAddCameraNode(H3DRootNode, "Camera", _forwardPipeRes);
	_reflectionCam = h3dAddCameraNode(H3DRootNode, "ReflectionCam", _forwardPipeRes);
	h3dSetNodeParamI(_cam, H3DCamera::OccCullingI, 1);
	h3dSetNodeParamI(_reflectionCam, H3DCamera::OccCullingI, 1);
	h3dSetNodeParamI(_reflectionCam, H3DCamera::OutTexResI, _waterplaneTexRes);

	//// Customize post processing effects
	//H3DNode matRes = h3dFindResource(H3DResTypes::Material, "pipelines/postHDR.material.xml");
	//h3dSetMaterialUniform(matRes, "hdrExposure", 2.5f, 0, 0, 0);
	//h3dSetMaterialUniform(matRes, "hdrBrightThres", 0.5f, 0, 0, 0);
	//h3dSetMaterialUniform(matRes, "hdrBrightOffset", 0.08f, 0, 0, 0);
	
	//Add sky
	H3DNode sky = h3dAddNodes(H3DRootNode, skyBoxRes);
	h3dSetNodeTransform(sky, 0, 150, 0,
		0, 90, 0, 
		310, 400, 310);
	
	// Create initial scenery
	_bodyCount = 0;
	_world = new World(this, _physics);

	for(int i = 0; i < ::Values::NUMBER_OF_ENEMIES; i++) {
		_enemies.push_back(new Enemy(this, _physics, H3DRootNode, _fontMatRes, i));
	}

	_player = new Player(this, _physics, _cam, 0);
	_player->addReflectionCamera(_reflectionCam);

	_airplane = new Airplane();
	return true;
}

/**
 * Actually the loop is in Start.main()
 * @param timeStep the time since the last rendering in millisecs
 */
void Game::update(const float timeStep, const float fps) {
	_runtime += timeStep;

	// update physics
	_physics->update(timeStep);
	// update player
	_player->update(timeStep);
	//update airplane
	_airplane->update(timeStep);
	//update world and entities in it
	_world->update(timeStep);

	//update enemies
	for (unsigned int i = 0; i < _enemies.size(); i++) {
		_enemies[i]->update(_cam, _player, timeStep);
	}

	//give time information to water shader
	H3DNode waterRes = h3dFindResource(H3DResTypes::Material, "models/waterplane/waterplane.material.xml");
	//if(! waterRes) std::cout << "no waterplane in Game found...\n";
	//else h3dSetMaterialUniform(waterRes, "TIME_FROM_INIT", _runtime, 0,0,0);
	
	// Show crosshair
	h3dShowOverlay(
		 0.484f, 0.48f, 0, 1,    // |
		 0.484f, 0.52f, 0, 0,    // |_
		 0.516f, 0.52f, 1, 0,    // _| 
		 0.516f, 0.48f, 1, 1,    // `|
		//colors
        1, 1, 1, 1, 
		_crossHairMatRes, 7);

	//display stats
	displayStats(timeStep, fps);

	displayHUD();

		
	//do reflections
	if(! _debugViewMode) {
		//h3dSetupViewport(20, 20, 120, 120, true);
		h3dRender(_reflectionCam);
	}

	// Render scene normally
	h3dSetupViewport(0, 0, _windowWidth, _windowHeight, true);
	h3dRender(_cam);

	// Finish rendering of frame
	h3dFinalizeFrame();

	// Remove all overlays
	h3dClearOverlays();

	// Write all messages to log file
	h3dutDumpMessages();
}

/**
 * display player information like health, current weapon or number of bullets left.
 */
void Game::displayHUD() const {
	std::ostringstream health, bullets;
	health << _player->getHealth() << "%";
	const char* weaponName = _player->getCurrentWeaponName();
	//TODO implement ammounition counting and fetching bullet packs also
	bullets << "Ammo "<< "infinite";

	h3dutShowText(health.str().c_str(), 0.00f, 0.95f,   0.04f, 1, 1, 1, _fontMatRes, 5);
	h3dutShowText(weaponName,           0.30f, 0.95f,   0.04f, 1, 1, 1, _fontMatRes, 5);
	h3dutShowText(bullets.str().c_str(),0.60f, 0.95f,   0.04f, 1, 1, 1, _fontMatRes, 5);
}

/**
 * Display game statistics like triangle count or fps.
 * @param timeStep the time delta that shall be rendered
 */
void Game::displayStats(const float timeStep, const float fps) const{
	int triangle = (int) h3dGetStat(H3DStats::TriCount, true);
	std::ostringstream stat1, stat2;

	float tx, ty, tz;
	float rx, ry, rz;
	float sx, sy, sz;
	h3dGetNodeTransform(_cam, &tx, &ty, &tz, &rx, &ry, &rz, &sx, &sy, &sz);
	Vec3 pos(tx, ty, tz);
	stat1 << "triangles: " << triangle << " fps:  " << fps;
	stat2 << "Position: " << pos.toString().c_str();
	h3dutShowText(stat1.str().c_str(), 0.01f, 0.015f,   0.02f, 1, 1, 1, _fontMatRes, 5);
	h3dutShowText(stat2.str().c_str(), 0.01f, 0.000f,   0.02f, 1, 1, 1, _fontMatRes, 5);
}

/**
 * resets a recently killed enemy until all enemies are killed
 */
void Game::recycleEnemy(Enemy* oldEnemy) {
	_bodyCount++;
	oldEnemy->reset((_bodyCount < ::Values::MAX_BODY_COUNT));
}

/**
 * @returns true if the number of shot enemies has not been reached
 */
bool Game::createMoreEnemies()  const{
	return (_bodyCount + ::Values::NUMBER_OF_ENEMIES) <= ::Values::MAX_BODY_COUNT;
}

/**
 * returns a vector of enemies
 */
std::vector<Enemy*> Game::getEnemies() const {
	return _enemies;
}

/**
 *
 * @return the only airplane that this game has got
 */
Airplane* Game::getAirplane() const {
	return _airplane;
}

/**
 *
 * @return the material ressource of the water reflection
 */
H3DRes Game::getReflectionTexture() const {
	return _waterplaneTexRes;
}

bool Game::hasPlayerAtLevelBounds(const Vec3 position) const {
	return _world->testInboundXZ(position);
}

////////////////////////////////////////////////
// static methods //////////////////////////////
////////////////////////////////////////////////

/**
 * animates particles of the given emitter
 * @param emitter the emitter to be animated
 * @param timeStep the time delta that shall be rendered
 */
void Game::advanceEmitter(H3DNode emitter, const float timeStep) {
	// Animate particle systems (several emitters in a group node)
	unsigned int cnt = h3dFindNodes(emitter, "", H3DNodeTypes::Emitter);

	for(unsigned int i = 0; i < cnt; ++i) {
		h3dAdvanceEmitterTime(h3dGetNodeFindResult(i), timeStep);
	}
}

/**
 * Adds light to the world in a compact manner.
 * @param px the position on the x axis relative to the world
 * @param py the position on the y axis relative to the world
 * @param pz the position on the z axis relative to the world
 * @param rx the rotation on the x axis relative to the world
 * @param ry the rotation on the y axis relative to the world
 * @param rz the rotation on the z axis relative to the world
 * @param length the length of the light cone from the light source away
 * @param fov the field of view, an angle that defines the reflection
 * @param shadowCount the number of shadows for the shadow context
 * @param r red color part of the light; from 0.0 to 1.0
 * @param g green color part of the light; from 0.0 to 1.0
 * @param b blue color part of the light; from 0.0 to 1.0
 * @return the created light
 */
H3DNode Game::addLight(H3DNode parent, const char* name,
		 const float px, const float py, const float pz, 
		 const float rx, const float ry, const float rz, 
		 const int length, const int fov, const int shadowCount, 
		 const float r, const float g, const float b) {

/* And God said, Let there be light and there was light.
  And God saw the light, that it was good: and God divided the LIGHTING from the SHADOWMAP.
*/
	H3DNode light = h3dAddLightNode(parent, name, 0, "LIGHTING", "SHADOWMAP");
	h3dSetNodeTransform(light, 
		px, py, pz, rx, ry, rz, 1, 1, 1);
	h3dSetNodeParamF(light, H3DLight::RadiusF, 0, length);
	h3dSetNodeParamF(light, H3DLight::FovF, 0, fov);
	h3dSetNodeParamI(light, H3DLight::ShadowMapCountI, shadowCount);
	h3dSetNodeParamF(light, H3DLight::ShadowSplitLambdaF, 0, 0.3f);
	h3dSetNodeParamF(light, H3DLight::ShadowMapBiasF, 0, 0.0005f);
	h3dSetNodeParamF(light, H3DLight::ColorF3, 0, r);
	h3dSetNodeParamF(light, H3DLight::ColorF3, 1, g);
	h3dSetNodeParamF(light, H3DLight::ColorF3, 2, b);
	return light;
}

////////////////////////////////////////////////
// application methods
////////////////////////////////////////////////

/**
 *
 */
void Game::resize(int width, int height ) {
	/*if(width > 1024) width = 1024;
	if(height > 768) width = 768;*/
	// Resize viewport
	h3dSetupViewport(0, 0, width, height, true);
	
	// Set virtual camera parameters
	h3dSetupCameraView(_cam, 45.0f, (float)width / height, 0.1f, 1000.0f);
	_windowWidth = width;
	_windowHeight = height;
}

/**
 *
 */
void Game::release() {
	// Release engine
	h3dRelease();
}

/**
 * sends mouse click events to the player
 * @param button which button is clicked
 * @param action true when clicked, false when released
 */
void Game::mouseButtonEvent(int button, bool action) {
	_player->setMouseButtonInput(button, action);
}

/**
 * sends keyboard events to the player
 */
void Game::keyStateChange(int key, bool state ) {
	_player->setKeyboardInput(key, state);
}

/**
 * sends mouse motion to the player
 */
void Game::mouseMoveEvent(float dX, float dY) {
	_player->setMouseInput(dX, dY);
}