#ifndef GAME_H_INCLUDED
#define GAME_H_INCLUDED

//Includes
#include <vector>
#include <sstream>
#include <stdexcept>

#include "Horde3D.h"
#include "Airplane.h"
#include "IShootable.h"

// Forward declarations
class World;
class Player;
class Enemy;
class Physics;
class IShootable;

//see below for ::Values-Struct
///////////////////////////////

/**
 * This class provides the start point of all graphics and physics simulation.
 * Although there is very little actual content added, this class is responsible
 * for general asset and object creation (Enemy, Player, World etc), and rendering as well
 * utilizing <pre>update(timeStep)</pre>.
 *
 * Please note that the given keyboard and mouse input is directed to Player.
 */
class Game {
private:
	World* _world;
	Player* _player;
	Physics* _physics;
	Airplane* _airplane;
	std::vector<Enemy*> _enemies;
	int _numberOfEnemies, _bodyCount, _statMode;
	bool               _debugViewMode, _wireframeMode, _isSplashScreenDeactivated;
	float              _runtime, _windowWidth, _windowHeight;
	
	// Engine objects
	H3DRes             _fontMatRes, _panelMatRes, _crossHairMatRes, _weaponMatRes, _waterplaneTexRes;
	H3DRes             _pipeRes, _hdrPipeRes, _forwardPipeRes, _reflectionPipeRes;
	H3DNode            _splashCam, _cam, _reflectionCam, _weapon;
public:
	Game();
	~Game();

	void update(const float timeStep, const float fps);
	bool init();
	void resize(int width, int height);
	void release();
	void recycleEnemy(Enemy* oldEnemy);
	std::vector<Enemy*> getEnemies() const;
	Airplane* getAirplane() const;
	H3DRes getReflectionTexture() const;
	bool hasPlayerAtLevelBounds(const Vec3 playerPosition) const;
	void keyStateChange(int key, bool state);
	void mouseMoveEvent(float dX, float dY);
	void mouseButtonEvent(int button, bool action);
	//static methods
	static void advanceEmitter(H3DNode emitter, const float timeStep);
	static H3DNode addLight(H3DNode parent, const char* name,
		 const float px, const float py, const float pz, 
		 const float rx, const float ry, const float rz, 
		 const int length, const int fov, const int shadowCount, 
		 const float r, const float g, const float b);
private:
	void displayHUD() const;
	void displayStats(const float timeStep, const float fps) const;
	bool createMoreEnemies() const;
};


/**
 * Values provides a means of bundled constants that are used throughout the game.
 */
struct Values {
	/**
	 * holds a string of the directory that holds the graphics content. It is relative to the start directory
	 */
	static const char*       CONTENT_DIR;
	/**
	 * provides the player's speed without running in Km/h. Running is 3 times faster
	 */
	static const float       WALKING_SPEED_KM_H;
	/**
	 * provides the airplane's speed in Km/h
	 */
	static const float       AIRPLANE_SPEED_KM_H;
	/**
	 * provides a single bullet's speed in m/s
	 */
	static const float       BULLET_SPEED_M_S;
	/**
	 * this is the player's maximum distance to hit IShootables
	 */
	static const float       SHOOTING_DISTANCE;
	/**
	 * provides the start health of Enemies and the Player
	 */
	static const int         MAX_HEALTH;
	/**
	 * Provides a scaleing factor for physics and graphics World
	 */
	static const int         WORLD_SCALE;
	/**
	 * provides the number of Enemies to be shot until no Enemies are respawned anymore
	 */
	static const int         MAX_BODY_COUNT;
	/**
	 * provides the number of concurrent Enemies
	 */
	static const int         NUMBER_OF_ENEMIES;
	/**
	 * provides a recoil factor for the chaingun weapon
	 */
	static const float       CHAINGUN_RECOIL;
	/**
	 * provides a recoil factor for the bolter weapon
	 */
	static const float       BOLTER_RECOIL;
};

#endif