#ifndef ISHOOTABLE_H_INCLUDED
#define ISHOOTABLE_H_INCLUDED

#include <btBulletDynamicsCommon.h>
#include "Horde3D.h"

#include "Physics.h"
#include "Game.h"

//forward declarations
class Game;
class Physics;

/**
 * This abstract class provides a possible interface to the player to hit with bullets. 
 * All implementing subclasses share the fact of a certain health state and the capability
 * to die by being hit with a bullet from the player.
 */
class IShootable {
protected:
	/**
	 * the id is used for general distinguation of objects of the same class
	 */
	int _id;
	/**
	 * After the health has reached zero this object
	 * is considered to be finally shot. the subclass may implement behaviour that may
	 * react to this state.
	 */
	int _health;
	/**
	 * The body is used to determine the collision with the bullet
	 */
	btRigidBody* _body;
	/**
	 * the parent may be used to attach this object to the scenery tree
	 */
	H3DNode _parent;
	/**
	 * accomodates the game as backreference
	 */
	Game* _game;
	/**
	 * accomodates physics since IShootables are always subject to collision detection
	 */
	Physics* _physics;
	/**
	 * the name of the shootable object
	 */
	const char* _name;
public:
	/**
	 * call this method to damage this IShootable. Hitting an instance would render in a 
	 * decreasing health until death ocurrs.
	 */
	virtual void hit() = 0; //pure virtual function

	/**
	 * intializes an IShootable 
	 * @param game
	 * @param parent
	 * @param physics
	 * @param id
	 */
	IShootable(Game* game, H3DNode parent, Physics* physics, const int id, const char* name)
		: _game(game), _parent(parent), _physics(physics), _id(id), _name(name) { }

	/**
	 * returns the id of this object
	 * @return the id
	 */
	int getID() {
		return _id;
	}

	/**
	 * returns the health of this object.
	 * @return the health of this object
	 */
	int getHealth() {
		return _health;
	}

	/**
	 * returns the physics body of this shootable object
	 * @return the body
	 */
	btRigidBody* getBody() {
		return _body;
	}

	/**
	 * returns the name of this shootable object
	 * @return the name
	 */
	const char* getName() {
		return _name;
	}
};

#endif