//includes
#include <iostream>
#include <vector>
#include <btBulletDynamicsCommon.h>

#include "Physics.h"
#include "Vec3.h"
#include "IShootable.h"
#include "Enemy.h"

/**
 * RayCastCallback provides a means of transportation of the hit point from a body to the world.
 */
class RayCastCallback : public btCollisionWorld::ClosestRayResultCallback {
	public:
		RayCastCallback(btCollisionObject* ignoredObject, const btVector3& from, const btVector3& to)
			: btCollisionWorld::ClosestRayResultCallback(from, to), ignoredObject(ignoredObject) {}
		virtual btScalar addSingleResult(btCollisionWorld::LocalRayResult &result, bool normal) {
			if (result.m_collisionObject == ignoredObject)
				return 1.0f;
			return btCollisionWorld::ClosestRayResultCallback::addSingleResult(result, normal);
		}
	private:
		btCollisionObject* ignoredObject;
};

/**
 * Constructs the physics world
 */
Physics::Physics() {
	// Build the broadphase
	int maxProxies = 1024;
	//world size is 5 Km in every direction
	btVector3 worldAabbMin(-5000,-5000,-5000);
	btVector3 worldAabbMax(5000,5000,5000);
	_broadphase = new btAxisSweep3(worldAabbMin,worldAabbMax,maxProxies);

	// Set up the collision configuration and dispatcher
	_collisionConfiguration = new btDefaultCollisionConfiguration();
	_dispatcher = new btCollisionDispatcher(_collisionConfiguration);

	// The actual physics solver
	// objects to interact properly, gravity, game logic supplied forces, collisions, and hinge constraints
	_solver = new btSequentialImpulseConstraintSolver;

	// The world.
	_dynamicsWorld = new btDiscreteDynamicsWorld(_dispatcher, _broadphase, _solver, _collisionConfiguration);
	//define Y axis as UP, gravity
	_dynamicsWorld->setGravity(btVector3(0,-30,0));
}

Physics::~Physics() {
	// Clean up behind ourselves like good little programmers
	//remove bodies, one by one
	for(unsigned int i = 0; i < _nonShootableBodies.size(); i++) {
		_dynamicsWorld->removeRigidBody(_nonShootableBodies[i]);
	}
	for(unsigned int i = 0; i < _shootableBodies.size(); i++) {
		if(_shootableBodies[i]->getID() >= 0) {
			_dynamicsWorld->removeRigidBody(_shootableBodies[i]->getBody());
		}
	}
	_nonShootableBodies.clear();
	_shootableBodies.clear();

	delete _dynamicsWorld;
	delete _solver;
	delete _collisionConfiguration;
	delete _dispatcher;
	delete _broadphase;
}

/**
 * adds a body to the world. Bodies added by this method are not subject to bullet collision calculation.
 * @see IShootable
 * @see addShootable()
 */
void Physics::addNonShootable(btRigidBody* body) {
	//TODO check if body exists already in world
	_nonShootableBodies.push_back(body);
	_dynamicsWorld->addRigidBody(body);
}

/**
 * adds a body to the world that are subject to bullet collision calculation.
 * @see addNonShootable()
 */
void Physics::addShootable(IShootable* shootable) {
	_shootableBodies.push_back(shootable);
	_dynamicsWorld->addRigidBody(shootable->getBody());
}

/**
 * Removes a shootable that has been added by addShootable(IShootable* shootable).
 * This is needed to avoid 
 * @param shootable the shootable object that should be removed
 * @returns true if 
 */
bool Physics::removeShootable(IShootable* shootable) {
	for(unsigned int i = 0; i < _shootableBodies.size(); i++) {
		if(_shootableBodies[i] == shootable) {
			_dynamicsWorld->removeRigidBody(shootable->getBody());
			_shootableBodies.erase(_shootableBodies.begin() + i);
			return true;
		}
	}
	return false;
}

/**
 * updates the world physics
 */
void Physics::update(float timeStep) {
	_dynamicsWorld->stepSimulation(timeStep);
}

//TODO create converter method for vec3->btvector3 or something

/**
 * conducts a collision test to the physics world
 * @return true if the world has been hit. the resulting point is then written to collisionPoint
 */
bool Physics::castRay(const Vec3& from, const Vec3& to, Vec3& collisionPoint, btRigidBody* ignore) {
	btVector3 fromBt(from.x, from.y, from.z);
	btVector3 toBt(to.x, to.y, to.z);
	
	// Perform ray cast
	RayCastCallback callback(ignore, fromBt, toBt);
	_dynamicsWorld->rayTest(fromBt, toBt, callback);
	// Copy resulting point
	if (callback.hasHit()) {
		collisionPoint = Vec3(
			callback.m_hitPointWorld.x(),
			callback.m_hitPointWorld.y(),
			callback.m_hitPointWorld.z());
		return true;
	}
	else
		return false;
}

/**
 * shoot a bullet or laser ray and see if someone has been hit
 * @return true if a non-world body has been hit
 */
IShootable* Physics::castGunRay(const Vec3& from, const Vec3& to, Vec3& collisionPoint, btRigidBody* ignore) {
	btVector3 fromBt(from.x, from.y, from.z);
	btVector3 toBt(to.x, to.y, to.z);
	
	// Perform ray cast
	RayCastCallback callback(ignore, fromBt, toBt);
	_dynamicsWorld->rayTest(fromBt, toBt, callback);

	// the bullet hits an enemy and not the world 
	if (callback.hasHit()) {
		IShootable* shootable = NULL;

		for(unsigned int i = 0; i < _shootableBodies.size(); i++) {
			//test if the collided object is the body of an IShootable
			if(callback.m_collisionObject == _shootableBodies[i]->getBody()) {
				shootable = _shootableBodies[i];
				break;
			}
		}
		//we have hit something that cannot be hit, f.e. world, cars, etc
		if(shootable == NULL) return NULL;

		//set point where collision has happened
		collisionPoint = Vec3(
			callback.m_hitPointWorld.x(),
			callback.m_hitPointWorld.y(),
			callback.m_hitPointWorld.z());
		return shootable;
	}
	return NULL;
}

/**
 * @return the physics world
 */
btCollisionWorld* Physics::getWorld() {
	return _dynamicsWorld;
}