#if defined(_MSC_VER) || defined(_WINDOWS_) || defined(_WIN32)
	#include "glfw.h"
#else
	#include <GL/glfw.h>
#endif
#include <iostream>
#include <cstdio>
#include <vector>
#include <time.h>
#include <stdlib.h>

#include <btBulletDynamicsCommon.h>
#include "Horde3D.h"
#include "Horde3DUtils.h"

#include "CharacterCollisionCallback.h"
#include "Player.h"
#include "Physics.h"
#include "IShootable.h"

/**
 * constructs the Player. The player has currently no viewable body.
 * @param game the game is needed to give and take important calls
 * @param camera the camera frustrum represents the player's view and is changed by the user's input
 * @param physics the physics are needed to conduct physics simulation
 * @param id an identifier
 * @see IShootable
 */
Player::Player(Game* game, Physics* physics, H3DNode camera, const int id)
		: IShootable(game, camera, physics, id, "Player") {

	//init values
	MOVE_LEFT =  0x0000001;
	MOVE_BACK =  0x0000010;
	MOVE_RIGHT = 0x0000100;
	MOVE_FWD =   0x0001000;
	MOVE_RUN =   0x0010000;
	FIRE_GUN =   0x0100000;
	MOVE_JUMP =  0x1000000;
	_rx = _ry = 0;
	_keys = 0x000000;
	_debugMode = false;
	_wireframeMode = false;
	_timeSinceLastBullet = 0;
	_health = ::Values::MAX_HEALTH;
	srand(time(NULL)); //needed for gunfire rotation
	
	//load weapons ressources
	H3DRes weaponChaingunRes = h3dAddResource(H3DResTypes::SceneGraph, "models/chaingun/chaingun.scene.xml", 0);
	H3DRes weaponBolterRes = h3dAddResource(H3DResTypes::SceneGraph, "models/bolter/bolter.scene.xml", 0);
	H3DRes weaponBolterAnim = h3dAddResource(H3DResTypes::Animation, "models/bolter/bolter.anim", 0);
	_bulletRes = h3dAddResource(H3DResTypes::SceneGraph, "models/bullet/bullet.scene.xml", 0);
	H3DRes gunfireRes = h3dAddResource(H3DResTypes::SceneGraph, "models/gunfire/gunfire.scene.xml", 0);
	
	bool loaded = h3dutLoadResourcesFromDisk(::Values::CONTENT_DIR);
	if(! loaded) std::cout << "could not load player ressources" << std::endl;

	_weaponChainGun = h3dAddNodes(_parent, weaponChaingunRes);
	_weaponBolter = h3dAddNodes(_parent, weaponBolterRes);
	h3dSetupModelAnimStage(_weaponBolter, 0, weaponBolterAnim, 0, "", true);
	_gunfire = h3dAddNodes(_parent, gunfireRes);
	_gunFireLight = Game::addLight(_parent, "GunFireLight", 
		0.13f, -0.1f, -0.6f,  //position
		0, 20, 0,             //rotation
		7, 80, 3,             //length, fov, shadow count
		1, 0.9f, 0.3f);       //color
	
	//deactive gun fire now and active it when the trigger is pulled
	h3dSetNodeActivation(_gunfire, false);
	h3dSetNodeActivation(_gunFireLight, false);

	h3dSetNodeTransform(_weaponChainGun, 0.35f, -0.45f, -1.0f,   5, 120, 10,  0.10f, 0.10f, 0.10f);
	h3dSetNodeTransform(_weaponBolter,   0.21f, -0.12f, -0.5f,   5, 110, 10,  0.08f, 0.08f, 0.08f);

	h3dSetNodeTransform(_gunfire, 
		0.1f, -0.10f, -0.5,   //position
		0, 55, 0,            //rotate // rz is set when the trigger is pulled
		0.02f, 0.02f, 0.02f); //scale

	_currentWeaponType = UNSET_WEAPON;
	//set weapon
	changeWeapon(BOLTER);

	// Create capsule shape
	_bodyRadius = 0.25f;
	_levelBelowCapsule = 0.20f;
	float bodyHeight = 1.86f;
	_capsuleHeight = bodyHeight - _levelBelowCapsule - (_bodyRadius * 2);
	_capsule = new btCapsuleShape(_bodyRadius, _capsuleHeight); 
	
	// Create physics body
	_groundMotionState = new btDefaultMotionState(btTransform(
			btQuaternion(0,0,0,1),
			btVector3(-100, -8.0f, 0)) // Start position of the player
	);
	
	// Create player body with 90 kg
	_body = new btRigidBody(90, _groundMotionState, _capsule);
	_physics->addShootable(this); //deconst-delete is managed by physics
}

Player::~Player() {
	//TODO clean up that mess, boy
}

/**
 * 
 * @param2
 */
void Player::addReflectionCamera(H3DNode cam2) {
	_cam2 = cam2;
}

/**
 * applies the changes made to the player's view. Changes ocurr from both player input
 * and physics, that is: gravitation and so on.
 * @param timeStep
 */
void Player::update(const float timeStep) {
	//render shot bullet or remove it for efficency's sake
	for(unsigned int i = 0; i < _shotBullets.size(); i++) {
		if(_shotBullets[i]->isActive()) {
			_shotBullets[i]->update(timeStep);
		}
		else {
			delete _shotBullets[i];
			_shotBullets.erase(_shotBullets.begin() + i);
		}
	}

	// Reset rotation of the body (so that the player will not fall)
	btTransform transform = _body->getCenterOfMassTransform();
	transform.setRotation(btQuaternion(0, 0, 0, 1));
	_body->setCenterOfMassTransform(transform);
	
	_body->setLinearVelocity(_body->getLinearVelocity() / 2.0f);
	_body->setAngularVelocity(btVector3(0,0,0));
	
	// Cast a ray from the bottom of the capsule 1m down
	Vec3 currentPosition = getPosition();
	Vec3 rollbackPosition = currentPosition;
	Vec3 collisionPoint;	
	bool onground = castRayAndCorrectPosition(currentPosition, collisionPoint);
	
	float vx = 0;
	float vy = 0;
	float vz = 0;
	float runMultiplier = 1;
	bool mouseClick = false;
	if(_keys & MOVE_FWD)   vz -= 1;
	if(_keys & MOVE_LEFT)  vx -= 1;
	if(_keys & MOVE_BACK)  vz += 1;
	if(_keys & MOVE_RIGHT) vx += 1;
	if(_keys & MOVE_RUN)   runMultiplier *= 3;
	if(_keys & FIRE_GUN)   mouseClick = true;
	if(_keys & MOVE_JUMP)  vy += 2; //jump

	//fire if weapon is ready and damage the enemey
	if(mouseClick) {
		//animate bullet round in magazin
		_animationTime = _animationTime + (60 * timeStep);
		h3dSetModelAnimParams(_weaponBolter, 0, _animationTime, 1.0);
		h3dSetModelAnimParams(_weaponChainGun, 0, _animationTime, 1.0);

		if(_timeSinceLastBullet > (_curWeaponRateOfFire / 1000.0f)) {
			//it is rate of fire time: FIRE!
			addGunFire(timeStep);

			IShootable* hitEnemy = fireGun(currentPosition, _rx, _ry);
			if(hitEnemy != NULL) {
				hitEnemy->hit();
			}

			//let it be {1,-1,0} for positive, negative or no recoil
			float randX = (rand() % 3) - 1;
			float randY = (rand() % 3) - 1;
			//add some recoil so the gun is not static; no perfect shooting for steady fire
			_rx += randX / 1.222f * _weaponRecoilFactor; //less X recoil than on Y
			_ry += randY / 2.000f * _weaponRecoilFactor; //so the gun acts like moving upwards
		}
		else  {
			//weapon not ready. add millisecs to counter
			_timeSinceLastBullet += clock() * CLOCKS_PER_SEC / 1000.0f;
			//no bullet (fireGun() takes care!)
			//no gunfire
			removeGunFire();
		}
	}
	else {
		//reset trigger counter
		_timeSinceLastBullet = 0;
		removeGunFire();
	}

	Vec3 speed(vx, vy, vz);
	// Don't let the player walk faster when strafing *and* walking forward
	if (speed.getLengthSquared() >= 2) {
		speed /= 1.4142f; //sqrt(2);
	}
	Vec3 speedRollback = speed;
	btTransform trans;

	// Take the intersection point (=the point at the feet of the player) and
	// set the position of the body so that this point is 0.2 m below the capsule.
	if(onground) {
		//if(velocity > _TOO_MUCH_FALLING) addDamage();
		// Reset linear velocity (stop after jumping)
		//_body->setLinearVelocity(btVector3(0,0,0));

		// Apply movement based on keyboard input
		speed *= (::Values::WALKING_SPEED_KM_H / 3.6 * timeStep) * runMultiplier;
		speed.rotateXZ(_ry);

		rollbackPosition = currentPosition;
		setPosition(currentPosition);
		//move the distance in speed
		currentPosition += speed;
	}
	else {
		_body->getMotionState()->getWorldTransform(trans);
		setPosition(Vec3(trans.getOrigin().x(), trans.getOrigin().y(), trans.getOrigin().z()));
	}

	//correct body collision detection to avoid to run thru things
		//check if other bodies collide
		btConvexShape *shape = (btConvexShape*)_body->getCollisionShape();
		btTransform from = _body->getCenterOfMassTransform();
		btTransform to = from;
		to.setOrigin(to.getOrigin() + btVector3(speed.x, 0, speed.z));
		// Check collisions
		CharacterCollisionCallback cb(_body, from.getOrigin(), to.getOrigin());
		_physics->getWorld()->convexSweepTest(shape, from, to, cb);
		if (cb.hasHit()) {
			// "currentPosition" does not reflect the actual position, reset it.
			//this time don't run, just walk
			speed *= (::Values::WALKING_SPEED_KM_H / 3.6 * timeStep) / 10.0f;
			currentPosition += speed;
			castRayAndCorrectPosition(currentPosition, collisionPoint);
			//currentPosition = rollbackPosition; //doesn't work. get stuck in the world
			setPosition(currentPosition);
		}
		else {
			setPosition(currentPosition);

			// Do a second raycast to ensure that we still after moving are
			// at the right distance to the terrain
			// Cast a ray from the bottom of the capsule down, again
			castRayAndCorrectPosition(currentPosition, collisionPoint);
			setPosition(currentPosition);
		}

    _body->getMotionState()->getWorldTransform(trans);
	// Update camera position/rotation to match the position of the body
	float playerOffset = ((btCapsuleShape*)_body->getCollisionShape())->getHalfHeight();
	// Set camera parameters
	h3dSetNodeTransform(_cam2, 
		currentPosition.x, -(currentPosition.y + playerOffset), currentPosition.z, 
		_rx, -_ry, 0, 
		1, 1, 1);

	h3dSetNodeTransform(_parent, 
		currentPosition.x, currentPosition.y + playerOffset, currentPosition.z, 
		_rx, _ry, 0, 
		1, 1, 1);

}

/**
 * corrects the vertical position of the physics body in respect to the terrain
 * @param calculatedCurrentPosition
 * @param collisionPoint
 * @param physics
 * @returns true if on ground, false if above or under the ground
 */
bool Player::castRayAndCorrectPosition(Vec3& calculatedCurrentPosition, Vec3& collisionPoint) {
	float capsuleBottom = (-_capsuleHeight / 2) - _bodyRadius;

	// Cast a ray from the bottom of the capsule down
	Vec3 from = calculatedCurrentPosition;
	Vec3 to = calculatedCurrentPosition;
	from.y += capsuleBottom + 1;
	to.y += capsuleBottom - 2;
	bool onground = _physics->castRay(from, to, collisionPoint, _body);

	// Reset the coordinate so that the capsule is above the collision point
	if (onground) {
		calculatedCurrentPosition = Vec3(
			calculatedCurrentPosition.x, 
			_capsuleHeight / 2 + _bodyRadius + _levelBelowCapsule + collisionPoint.y,
			calculatedCurrentPosition.z);
	}
	return onground;
}

/**
 * this method should be called from the enemy when he hits the player.
 * The health should be decreased in turn. this method is not implemented, yet.
 */
void Player::hit() {
	//TODO fillme
}

/**
 * sets the current player position in world space
 * @param newPosition
 */
void Player::setPosition(const Vec3& newPosition) {
	btTransform transform;
	transform = _body->getCenterOfMassTransform();
	transform.setOrigin(btVector3(newPosition.x, newPosition.y, newPosition.z));
	_body->setCenterOfMassTransform(transform);
}

/**
 * returns the current player position in world space
 * @return a Vec3 represention the current position
 */
Vec3 Player::getPosition() const {
	btTransform transform = _body->getCenterOfMassTransform();
	btVector3 pos = transform.getOrigin();
	return Vec3(pos.x(), pos.y(), pos.z());
}

/**
 * returns the name of the selected weapon
 */
const char* Player::getCurrentWeaponName() const {
	switch(_currentWeaponType) {
		case CHAINGUN:
			return "Chaingun";
		case BOLTER:
			return "Bolter";
		case UNSET_WEAPON:
		default:
			return "???";
	}
}

////////////////////////////////////////////
// weapon stuff
////////////////////////////////////////////

/**
 * simulates a players shot and find shootable objects in a raycast (like Explosive or Enemy)
 * @param currentPosition the player's position in world space
 * @param rx the player's X rotation in world space
 * @param ry the player's > rotation in world space
 * @return an IShootable if something got hit, NULL otherwise
 */
IShootable* Player::fireGun(const Vec3 currentPosition, float rx, float ry) {
	// Cast a ray from the bottom of the capsule 1m down, again
	Vec3 from = currentPosition;
	//shot a ray into viewing direction
	Vec3 to(0, 0, -::Values::SHOOTING_DISTANCE);

	to.rotateYZ(rx);
	to.rotateXZ(ry);

	Vec3 direction = to;
//std::printf("from: %s\nto:  %s\n", from.toString().c_str(), to.toString().c_str());	
	H3DNode bulletProxy = h3dAddGroupNode(H3DRootNode, "bulletProxy");

	float ntx, nty, ntz;
	float nrx, nry, nrz;
	float nsx, nsy, nsz;
	h3dGetNodeTransform(_parent, &ntx, &nty, &ntz, &nrx, &nry, &nrz, &nsx, &nsy, &nsz);

	h3dSetNodeTransform(bulletProxy,
		ntx, nty, ntz, //position
		rx, ry, 0,  //rotate
		0.2f, 0.2f, 0.2f); //scale

	to += from;
//std::printf("from: %s\nto:  %s\n", from.toString().c_str(), to.toString().c_str());
	Vec3 collisionPoint;
	IShootable* enemy = _physics->castGunRay(from, to, collisionPoint, _body);
	float shootingDistance = ::Values::SHOOTING_DISTANCE;
	if(enemy != NULL) {
		shootingDistance = collisionPoint.getEuclideanDistance(from);
	}
	Bullet* bullet = new Bullet(_physics, bulletProxy, _bulletRes, shootingDistance);

	_shotBullets.push_back(bullet);
	return enemy;
}

/**
 * change the current weapon against another one
 * this is done by node activation rather than node add/remove
 * @param newWeapon
 */
void Player::changeWeapon(Weapon newWeapon) {
	bool isActive = true;
	if(newWeapon == _currentWeaponType) return;

	switch(newWeapon) {
		case CHAINGUN:
			_curWeaponRateOfFire = 100.0f;// rounds/s
			_weaponRecoilFactor = ::Values::CHAINGUN_RECOIL;
			h3dSetNodeActivation(_weaponChainGun, isActive);
			h3dSetNodeActivation(_weaponBolter, ! isActive);
			//due to different weapon models, the gunfire needs to be adjusted
			h3dSetNodeTransform(_gunfire, 
				0.1f, -0.085f, -0.5,  //position
				0, 55, 0,             //rotate // rz is set when the trigger is pulled
				0.02f, 0.02f, 0.02f); //scale
			break;

		case BOLTER:
			_curWeaponRateOfFire = 16.66667f;// rounds/s
			_weaponRecoilFactor = ::Values::BOLTER_RECOIL;
			h3dSetNodeActivation(_weaponChainGun, ! isActive);
			h3dSetNodeActivation(_weaponBolter,     isActive);
			//due to different weapon models, the gunfire needs to be adjusted
			h3dSetNodeTransform(_gunfire, 
				0.13f, -0.10f, -0.62,  //position
				0, 40, 0,              //rotate // rz is set when the trigger is pulled
				0.013f, 0.02f, 0.02f); //scale
			break;
	}
	_currentWeaponType = newWeapon;
}

/**
 * @return the number of possible shots per time step
 * @deprecated please check if this method is any good
 */
int Player::calcShotsPerTimeStep(float timeStep) {
	float shots = timeStep / 1000 * _curWeaponRateOfFire;
	return (int)floor(shots) * 1;
}

/**
 * adds a rotated gunfire node to the view
 */
void Player::addGunFire(const float timeStep) {
	_timeSinceLastBullet -= _curWeaponRateOfFire;

	//may deliver values > 360 but H3D can cope with that
	int gunfireRotation = rand() % 360;
	//acquire current position
	float ntx, nty, ntz;
	float nrx, nry, nrz;
	float nsx, nsy, nsz;
	h3dGetNodeTransform(_gunfire, &ntx, &nty, &ntz, &nrx, &nry, &nrz, &nsx, &nsy, &nsz);
	//set the same values but with a rotated gunfire
	h3dSetNodeTransform(_gunfire, ntx, nty, ntz, nrx, nry, gunfireRotation, nsx, nsy, nsz);
	h3dSetNodeActivation(_gunfire, true);
	bool doLight = ((rand() % 3) + 1) > 2 ;
	h3dSetNodeActivation(_gunFireLight, doLight); //set it true or false

}

/**
 * removes the gun fire from view when the trigger is not pressed
 */
void Player::removeGunFire() {
	//deactivate gunfire
	h3dSetNodeActivation(_gunfire, false);
	h3dSetNodeActivation(_gunFireLight, false);
}

/////////////////////////////////////////////////////////
// external input stuff
/////////////////////////////////////////////////////////

/**
 * does actions according to a pressed key
 * @param key
 * @param pushed true when pressed, false when released
 */
void Player::setKeyboardInput(const int key, bool pushed) {
	switch(key) {
		case 'A':
			if (pushed)
				_keys |=  MOVE_LEFT;
			else
				_keys &= ~MOVE_LEFT;
			break;
		case 'S':
			if (pushed)
				_keys |=  MOVE_BACK;
			else
				_keys &= ~MOVE_BACK;
			break;
		case 'D':
			if (pushed)
				_keys |=  MOVE_RIGHT;
			else
				_keys &= ~MOVE_RIGHT;
			break;
		case 'W':
			if (pushed)
				_keys |=  MOVE_FWD;
			else
				_keys &= ~MOVE_FWD;
			break;
		case GLFW_KEY_LSHIFT:
			if (pushed)
				_keys |=  MOVE_RUN;
			else
				_keys &= ~MOVE_RUN;
			break;
		case GLFW_KEY_SPACE:
			if (pushed)
				_keys |=  MOVE_JUMP;
			else
				_keys &= ~MOVE_JUMP;
			break;
		case '1':
			if (! pushed) {
				changeWeapon(CHAINGUN);
			}
			break;
		case '2':
			if (! pushed) {
				changeWeapon(BOLTER);
			}
			break;
		case GLFW_KEY_F2:
			if (!pushed) {
				h3dSetOption( H3DOptions::DebugViewMode, _debugMode ? 1.0f : 0.0f );
				_debugMode = !_debugMode;
			}
			break;
		case GLFW_KEY_F3:
			if (!pushed) {
				h3dSetOption( H3DOptions::WireframeMode, _wireframeMode ? 1.0f : 0.0f );
				_wireframeMode = !_wireframeMode;
			}
			break;
	}
}

/**
 * adjusts the viewing angle which is used for the camera
 * according to the mouse movement
 */
void Player::setMouseInput(float dX, float dY) {
	// Look left/right
	_ry -= dX / 100 * 30;
	
	// Look up/down but only in a limited range
	_rx += dY / 100 * 30;
	if( _rx > 90 ) _rx = 90; 
	if( _rx < -90 ) _rx = -90;
}

/**
 * fires a gun when the left mouse button is clicked
 */
void Player::setMouseButtonInput(int button, bool action) {
	if (button == GLFW_MOUSE_BUTTON_1 && action)
		_keys |=  FIRE_GUN;
	else
		_keys &= ~FIRE_GUN;
}