#ifndef VEC3_H_INCLUDED
#define VEC3_H_INCLUDED

#include <iostream>
#include <sstream>
#include <string>
#include <math.h>
#include <cmath>
#include "utMath.h"

/**
 * This class provides a triple vector for (X,Y,Z) and matching vector calculations as well, that are
 * mostly useful for 3D rotation calculations.
 * The vector components are internally stored as floats and are directly available for read/write access.
 */
class Vec3 {
public:
	/**
	 * read/write access the X component
	 */
	float x;
	/**
	 * read/write access the Y component
	 */
	float y;
	/**
	 * read/write access the Z component 
	 */
	float z;
public:
	/**
	 * constructs a Vec3 with the given vector components. Components may be occluded and are seth then as zero.
	 * @param the x value to be set
	 * @param the y value to be set
	 * @param the z value to be set
	 */
	Vec3(const float x=0.0f, const float y=0.0f, const float z=0.0f)
		: x(x), y(y), z(z) {
	}
	/**
	 * A copy-constructor for a similar vector class Vec3f from the Horde3D framework.
	 * @param v the vector whose components are to be copied
	 */
	Vec3(const Vec3f v)
		: x(v.x), y(v.y), z(v.z) {
	}
	~Vec3() {
	}
	
	//////////////////////////////////////////////
	// operator overloading
	//////////////////////////////////////////////

	/**
	 * this operator does a component-wise addition to this vector
	 * @param b the vector which is added to this vector
	 * @return this vector but updated
	 */
	Vec3 &operator+=(const Vec3 &b) {
		x += b.x;
		y += b.y;
		z += b.z;
		return *this;
	}

	/**
	 * this operator does a component-wise addition and returns a fresh vector, while
	 * this vector is left unchanged.
	 * @param b the vector which is added to this vector's components
	 * @return a fresh vector
	 */
	Vec3 operator+(const Vec3 &b) const {
		return Vec3(x + b.x, y + b.y, z + b.z);
	}

	/**
	 * this operator does a component-wise substraction and returns a fresh vector, while
	 * this vector is left unchanged.
	 * @param b the vector which is substracted from this vector
	 * @return a fresh vector
	 */
	Vec3 operator-(const Vec3 &b) const {
		return Vec3(x - b.x, y - b.y, z - b.z);
	}

	/**
	 * this operator does a component-wise multiplication and returns a fresh vector, while
	 * this vector is left unchanged.
	 * @param b the vector which is multiplicated with this vector's components
	 * @return a fresh vector
	 */
	Vec3 operator*(const float b) const {
		return Vec3(x * b, y * b, z * b);
	}

	/**
	 * this operator does a component-wise multiplication to this vector.
	 * @param b the vector which is multiplicated with this vector's components
	 * @return this vector but updated
	 */
	Vec3 &operator*=(const float b) {
		x *= b;
		y *= b;
		z *= b;
		return *this;
	}

	/**
	 * this operator does a component-wise division to this vector.
	 * @param b the vector by which are this vector's components divided
	 * @return this vector but updated
	 */
	Vec3 &operator/=(const float b) {
		x /= b;
		y /= b;
		z /= b;
		return *this;
	}

	///////////////////////////////////////////
	// support methods
	///////////////////////////////////////////

	/**
	 * rotate this vector around the y axis
	 * @param the angle in degrees by which X and Z shall be rotated
	 */
	void rotateXZ(float ry) {
		float cosry = cos(degToRad(ry));
		float sinry = sin(degToRad(ry));
		float newz = (z * cosry) - (x * sinry);
		x = (z * sinry) + (x * cosry);
		z = newz;

	}

	/**
	 * rotate this vector around the z axis
	 * @param the angle in degrees by which X and Y shall be rotated
	 */
	void rotateXY(float rz) {
		float cosrz = cos(degToRad(rz));
		float sinrz = sin(degToRad(rz));
		float newy = (y * cosrz) - (x * sinrz);
		x = (y * sinrz) + (x * cosrz);
		y = newy;
	}

	/**
	 * rotate this vector around the x axis
	 * @param the angle in degrees by which Y and Z shall be rotated
	 */
	void rotateYZ(float rx) {
		float cosrx = cos(degToRad(rx));
		float sinrx = sin(degToRad(rx));
		float newz = (z * cosrx) + (y * sinrx);
		y = -(z * sinrx) + (y * cosrx);
		//Je größer der Winkel, desto weniger trägt das alte y zu dem neien y bei.
		//je größer der Winkel, desto mehr zieht das alte z (das nach unten(!) geklappt wird) von y ab.
		
		// mehr = sinus, weniger = cosinus. Für die ersten 90 Grad
		// Sinus: /--\__/, Cosinus: -\__/--
		//        ^^                ^^
		z = newz;
	}

	/**
	 * returns 
	 * @return the sum of all squared components
	 */
	float getLengthSquared() const {
		return (x * x) + (y * y) + (z * z);
	}

	/**
	 * returns the distance between this vector and another vector within the euclidean 3D space.
	 * @param pos2 another vector
	 * @return the distance between this vector and another vector. Please
	 *		note that this values is absolute.
	 */
	float getEuclideanDistance(const Vec3 pos2) const {
		float preResultX = (x - pos2.x)*(x - pos2.x);
		float preResultY = (y - pos2.y)*(y - pos2.y);
		float preResultZ = (z - pos2.z)*(z - pos2.z);
		return sqrt(preResultX + preResultY + preResultZ);
	}

	/**
	 * returns the angle between this vector and another vector (at most the player's position)
	 * @param the player's position
	 * @return the angle between both vectors in degree
	 */
	float getAngle(Vec3 playerPos) const {
		float dx = playerPos.x - x;
		float dz = playerPos.z - z;
		float result = atan2(dx, dz) * 180 / 3.1415f;
		return result;
	}

	/**
	 * converts this vector to a Horde3D vector for convenience purposes
	 * @return a fresh Horde3D vector with this vector's component values
	 */
	Vec3f toVec3f() const {
		return Vec3f(x,y,z);
	}

	/**
	 * returns a string with the current component values
	 * @return a string with the current component values
	 */
	std::string toString() {
		std::ostringstream result;
		result << "[ x: " << x << " y: " << y << " z: " << z << " ]";
		return result.str();
	}
};
#endif