#include <iostream>
#include <sstream>
#include <string>
#include <fstream>
#include <vector>
#include <time.h>
#include <stdlib.h>
#include <math.h>

#include <btBulletDynamicsCommon.h>
#include "Horde3D.h"
#include "Horde3DUtils.h"

#include "World.h"
#include "Physics.h"
#include "Vec3.h"
#include "Game.h" //delivers also constants
#include "Explosive.h"

/**
 * Constructs the world in respect of the physics and the graphics part.
 * @param game the game is needed to fetch needed data like reflection textures
 * @param physics the physics is needed to submit created physics objects
 */
World::World(Game* game, Physics* physics) {
	_debugMode = false;
	_waterAnimationTime = 0.0f;
	_game = game;
	_physics = physics;
	buildGraphicsWorld();
	buildPhysicsWorld();
}

/**
 * build the graphics part of the world. Included are
 * - landscape
 * - lighting
 * - miscellaneous models
 * - some particle emitters
 */
void World::buildGraphicsWorld() {
	// gfx world
	H3DRes envRes = h3dAddResource(H3DResTypes::SceneGraph, "models/world/world.scene.xml", 0 );
	// gfx objects
	//fixme H3DRes pickupRes = h3dAddResource(H3DResTypes::SceneGraph, "models/fordpickup/fordpickup.scene.xml", 0);
	H3DRes vulcanoRes = h3dAddResource(H3DResTypes::SceneGraph, "particles/vulcanoSmoke/vulcanoSmoke.scene.xml", 0);
	H3DRes houseRes = h3dAddResource(H3DResTypes::SceneGraph, "models/house/house.scene.xml", 0);
	H3DRes waterTowerRes = h3dAddResource(H3DResTypes::SceneGraph, "models/watertower/watertower.scene.xml", 0);
	_waterplaneRes = h3dAddResource(H3DResTypes::SceneGraph, "models/waterplane/watersurface.scene.xml", 0);
	H3DRes stopSignRes = h3dAddResource(H3DResTypes::SceneGraph, "models/signs/stop.scene.xml", 0);
	H3DRes treeRes = h3dAddResource(H3DResTypes::SceneGraph, "models/tree/tree.scene.xml", 0);
	H3DRes jeepRes = h3dAddResource(H3DResTypes::SceneGraph, "models/jeep/jeep.scene.xml", 0);
	H3DRes barrelRes = h3dAddResource(H3DResTypes::SceneGraph, "models/barrel/barrel.scene.xml", 0);
	
	H3DRes barrel2Res = h3dAddResource(H3DResTypes::SceneGraph, "models/drum/drum.scene.xml", 0);
	_crateRes = h3dAddResource(H3DResTypes::SceneGraph, "models/crate/woodencrate.scene.xml", 0);
	H3DRes tireRes = h3dAddResource(H3DResTypes::SceneGraph, "models/tire/tire.scene.xml", 0);
	
	H3DRes shipRes = h3dAddResource(H3DResTypes::SceneGraph, "models/ship/ship.scene.xml", 0);

	// Load resources
	bool loaded = h3dutLoadResourcesFromDisk(::Values::CONTENT_DIR);
	if(! loaded) std::cout << "could not load one or more world ressources" << std::endl;
	// Add environment
	H3DNode model = h3dAddNodes(H3DRootNode, envRes);
	
	//add objects to world
	h3dSetNodeTransform(model, 0, -10, 0, 0, 0, 0, ::Values::WORLD_SCALE, ::Values::WORLD_SCALE, ::Values::WORLD_SCALE);

	//add general light source
	//simpler form than all those parameters
	Game::addLight(H3DRootNode, "Light1", 200, 300, -170, //pos
		-30, 110, 0, //angle
		1900, 140, 4, //param
		1,1,1); // RGB color

	//Game::addLight(H3DRootNode, "Light2", -10, 200, -140, 
	//	-90, 0, 0,
	//	200, 120, 3,
	//	1,1,1);

	Game::addLight(H3DRootNode, "Vulcano", -84.5, 34.5, -134, 
		90, 0, 0,
		70, 120, 1,
		1, 0.3f, 0.5f);

	//add vulcano particles
	_vulcanoEmitter = h3dAddNodes(H3DRootNode, vulcanoRes);
	h3dSetNodeTransform(_vulcanoEmitter, -84.5, 34.5, -134, 
		90, 0, 0, 
		1, 1, 1);

	//exchange pond texture with camera reflection
	//at this point waterPlane must have been already set
	H3DNode waterplane = h3dAddNodes(H3DRootNode, _waterplaneRes);
	h3dSetNodeTransform(waterplane, -50, -25.8f, 70, 0, 0, 0, 35, 0.001, 35);
	//set water reflection texture
	H3DNode mesh = h3dGetNodeChild(waterplane, 0);
	std::cout << "found mesh " << mesh << std::endl;
	_waterplaneMeshMaterial = h3dGetNodeParamI(mesh, H3DMesh::MatResI);
	h3dSetMaterialUniform(_waterplaneMeshMaterial, "waterTime", 0.01f, 0.01f, 0.01f, 0.01f);

	std::cout << "found material " << _waterplaneMeshMaterial << std::endl;
	int idx = h3dFindResElem(_waterplaneMeshMaterial, H3DMatRes::SamplerElem, H3DMatRes::SampNameStr, "albedoMap");
	std::cout << "found index " << idx << std::endl;
	H3DRes texRes  = h3dGetResParamI(_waterplaneMeshMaterial, H3DMatRes::SamplerElem, idx, H3DMatRes::SampTexResI);
	std::cout << "found texres " << texRes << std::endl;
	h3dSetResParamI(_waterplaneMeshMaterial, H3DMatRes::SamplerElem, idx, H3DMatRes::SampTexResI, _game->getReflectionTexture());
	std::cout << "reflex res " << _game->getReflectionTexture() << std::endl;

	int idx2 = h3dFindResElem(_waterplaneMeshMaterial, H3DMatRes::SamplerElem, H3DMatRes::SampNameStr, "ambientMap");
	std::cout << "found index2 " << idx2 << std::endl;
	H3DRes texRes2  = h3dGetResParamI(_waterplaneMeshMaterial, H3DMatRes::SamplerElem, idx2, H3DMatRes::SampTexResI);
	std::cout << "found texres2 " << texRes2 << std::endl;
	h3dSetResParamI(_waterplaneMeshMaterial, H3DMatRes::SamplerElem, idx2, H3DMatRes::SampTexResI, _game->getReflectionTexture());
	std::cout << "reflex res2 " << _game->getReflectionTexture() << std::endl;
	
	//fixme H3DNode pickup = h3dAddNodes(H3DRootNode, pickupRes);
	//fixme h3dSetNodeTransform(pickup, 20, 10, 20, 0, 0, 0, 0.1,0.1,0.1);

	addModel(houseRes,      Vec3(-110, -4.8f, 21),  Vec3(0, 90, 0),  0.2);
	addModel(waterTowerRes, Vec3(-119, -9.8f, -14), Vec3(0, -90, 0), 10);
	addModel(stopSignRes,   Vec3(-69, -8.1f, 13),   Vec3(0, -90, 0), 0.25f);
	addModel(jeepRes,       Vec3(-81, -8.9f, 16),   Vec3(0, 71, 0),  3);
	addModel(_crateRes,      Vec3(-73, -9, 12),     Vec3(0, 71, 1),  1);
	addModel(barrel2Res,    Vec3(-55, -9, 12),      Vec3(0, 71, 1),  2);
	addModel(tireRes,       Vec3(-57, -9.8f, 12),   Vec3(0, 0, 90),  1);
	_ship = addModel(shipRes, Vec3(-85.7, -12.3f, 41.6f), Vec3(-13, -3, 3), 0.5);

	//add trees
	srand(time(NULL));
	for(int i = 0; i < 8; i++) {
		int rndx = rand() % 3;
		int rndz = (rand() % 65) + 10;
		addModel(treeRes, Vec3(-143 + rndx, -9.7f, 10 + rndz), Vec3(0, rndz + rndx, 0), 1.7f);
	}

	//add explosive barrels
	_models = std::vector<Explosive*>();
	Explosive* expl1 = new Explosive(_game, H3DRootNode, _physics, 0, barrelRes, Vec3(-75, -5.5f, -0.0f), Vec3(0, 0, 0), 0.33f);
	Explosive* expl2 = new Explosive(_game, H3DRootNode, _physics, 0, barrelRes, Vec3(-75, -7.5f, -0.5f), Vec3(0, 0, 0), 0.33f);
	Explosive* expl3 = new Explosive(_game, H3DRootNode, _physics, 0, barrel2Res, Vec3(-75, -8.0f, -1.0f), Vec3(0, 0, 0), 2.0f);
	_models.push_back(expl1);
	_models.push_back(expl2);
	_models.push_back(expl3);
}

/**
 * creates a H3DNode from a given resource on the root node
 * @param resource this resource is used to create the graphics scene node
 * @param position the position in world space
 * @param rotation the rotation in world space
 * @param scale this scale is applied on the X,Y, and Z axis, since it us pretty unusual to skew a model
 */
H3DNode World::addModel(H3DRes resource, Vec3 position, Vec3 rotation, float scale) {
	H3DNode model = h3dAddNodes(H3DRootNode, resource);
	h3dSetNodeTransform(model,
		position.x, position.y, position.z,
		rotation.x, rotation.y, rotation.z,
		scale, scale, scale);
	return model;
}

/**
 * checks whether the given position is inside the world.
 * @param position the position to be checked
 */
bool World::testInboundXZ(const Vec3 position) {
	return false;
}

/**
 * build the physics part from a wavefront data file
 * @param physics
 */
void World::buildPhysicsWorld() {
	// Open file for world coordinates
	std::string worldFilename("");
	worldFilename += Values::CONTENT_DIR;
	worldFilename += "/models/world/world.scene.obj";

	std::ifstream file(worldFilename.c_str(), std::ifstream::in);
	if (!file) {
		std::cout << "Could not open file " << worldFilename.c_str() << std::endl;
		return;
	}

	// The following code reads a line like "v 1.0 2.0 3.0" (no commas between values)
	// Read one character, like cout/cin
	char type;

	std::vector<Vec3> vertices;
	std::vector<int> indices;

	int vertexCount = 0;
	int indexCount= 0;

	//recreate stream 
	//istringstream iss = istringstream(text, istringstream::in);

	int vi = 0;
	int ii = 0;

	while(! file.eof()) {
		file >> type;
		//skip lines without useful content
		if (type == '#') { std::string line; std::getline(file, line); } //comments
		else if (type == 'm') { std::string line; std::getline(file, line); } // mtllib
		else if (type == 'u') { std::string line; std::getline(file, line); } // usemtl
		else if (type == 's') { std::string line; std::getline(file, line); } // s off
		// Whitespaces are skipped!
		else if (type == 'v') {
			float x, y, z;
			file >> x >> y >> z;
			vertices.push_back(Vec3(x, y, z) * ::Values::WORLD_SCALE);
			vi++;
		}
		else if (type == 'f') {
			int x, y, z;
			file >> x >> y >> z;
			indices.push_back(x);
			indices.push_back(y);
			indices.push_back(z);
			ii++;
		}
	}

	// Create trimesh for physics
	trimesh = new btTriangleMesh();
	// Insert world triangles here
	for (unsigned int i = 0; i < indices.size() / 3; i++) {
		//obj data starts with one
		int index0 = indices[i * 3 + 0] - 1;


		int index1 = indices[i * 3 + 1] - 1;
		int index2 = indices[i * 3 + 2] - 1;

		btVector3 vertex0(vertices[index0].x, vertices[index0].y, vertices[index0].z);
		btVector3 vertex1(vertices[index1].x, vertices[index1].y, vertices[index1].z);
		btVector3 vertex2(vertices[index2].x, vertices[index2].y, vertices[index2].z);

		trimesh->addTriangle(vertex0, vertex1, vertex2);
	}

	groundMotionState = new btDefaultMotionState(btTransform(
			btQuaternion(0,0,0,1),
			btVector3(0,-10,0))
	);

	// Create trimesh shape // shape => collision detection
	btBvhTriangleMeshShape *shape = new btBvhTriangleMeshShape(trimesh, true);
	// Create body with zero mass => not moving
	btRigidBody::btRigidBodyConstructionInfo groundBodyCI(0, groundMotionState, shape, btVector3(0,0,0));
	body = new btRigidBody(groundBodyCI); //Physics takes care of garbage collection for bullet objects
	_physics->addNonShootable(body);

	//TODO
	//build no-go-area to the world boundings
	addPhysicsBox(Vec3(233, 30, 2), Vec3(0,-8, 135), 0); //border south
	addPhysicsBox(Vec3(233, 30, 2), Vec3(0,-8,-150), 0); //border north
	addPhysicsBox(Vec3(2, 30, 250), Vec3(-120,-8,0), 0); //border west
	addPhysicsBox(Vec3(2, 30, 250), Vec3(170,-8,0), 0); //border east

	addPhysicsBox(Vec3(2, 2, 5), Vec3(-52.9f,-8, 11), 0); //car
	//addPhysicsBox(Vec3(2, 2, 5), Vec3(170,-8,0), 0); //boat
	addPhysicsBox(Vec3(4, 12, 5), Vec3(170,-13, 80), 0); //pond
	addPhysicsBox(Vec3(0.2f, 5, 0.2f), Vec3(49.9f,-7.4f,12.9f), 0); //stop sign
	addPhysicsBox(Vec3(2, 12, 5), Vec3(170,-8,0), 0); //water tower

	//house boxes
	addPhysicsBox(Vec3(0.3f, 0.3f, 2.3f), Vec3(-105.5f, -9.5f, 27.3f), 0); //step1
	addPhysicsBox(Vec3(0.3f, 0.3f, 2.3f), Vec3(-106.0f, -9.0f, 27.3f), 0); //step2
	addPhysicsBox(Vec3(0.3f, 0.3f, 2.3f), Vec3(-106.5f, -8.5f, 27.3f), 0); //step3
	addPhysicsBox(Vec3(10.f, 0.3f, 5), Vec3(-117, -8.5f, 23.0f), 0); //house platform
	//addPhysicsBox(Vec3(2, 2, 3), Vec3(170,-8,0), 0); //wall W
	//addPhysicsBox(Vec3(2, 2, 3), Vec3(170,-8,0), 0); //wall S
	//addPhysicsBox(Vec3(2, 2, 2), Vec3(170,-8,0), 0); //wall E
	//addPhysicsBox(Vec3(2, 2, 2), Vec3(170,-8,0), 0); //wall N

}

/**
 * puts separate bodies to the world to simulate solid bodies or create landscape borders
 * @param halfsize
 * @param position
 */
void World::addPhysicsBox(const Vec3 halfsize, const Vec3 position, const float rotationY) {
	btCollisionShape* shape = new btBoxShape(btVector3(halfsize.x, halfsize.y, halfsize.z));
	btMotionState* motionState = new btDefaultMotionState(btTransform(
			btQuaternion(0,0,0,1),
			btVector3(position.x, position.y, position.z)));
	const float ZERO_MASS = 0.0f;
	btRigidBody* wall = new btRigidBody(ZERO_MASS, motionState, shape);
	wall->setActivationState(WANTS_DEACTIVATION);
	if(_debugMode) {
		H3DNode theNode = h3dAddNodes(H3DRootNode, _crateRes);
		h3dSetNodeTransform(theNode, position.x, position.y, position.z,
			0,0, rotationY, 2 * halfsize.x, 2 * halfsize.y, 2 * halfsize.z);
	}

	_physics->addNonShootable(wall);
}

/**
 * updates the graphics state for world entities
 * @param timeStep the time since the last rendering in millisecs
 */
void World::update(const float timeStep) {
	for(unsigned int i = 0; i < _models.size(); i++) {
		Explosive* exp = _models[i];
		exp->update(timeStep);
	}
	// Animate particle systems (several emitters in a group node)
	Game::advanceEmitter(_vulcanoEmitter, timeStep);
	
	//animate pond water by shader calculations
	const float slownessFactor = 8.0f;
	_waterAnimationTime += timeStep / slownessFactor;
	float animationTime = tan(_waterAnimationTime);
	//std::cout << "sin = " << animationTime << std::endl;
	//waterTime corresponds to uniform in material.xml
	float notUsed = -1;
	h3dSetMaterialUniform(_waterplaneMeshMaterial, "waterTime", animationTime, notUsed, notUsed, notUsed);
	
	//let the small ship linger in the pond
	float ntx, nty, ntz;
	float nrx, nry, nrz;
	float nsx, nsy, nsz;
	h3dGetNodeTransform(_ship, &ntx, &nty, &ntz, &nrx, &nry, &nrz, &nsx, &nsy, &nsz);
	h3dSetNodeTransform(_ship, ntx, nty, ntz, animationTime * slownessFactor, nry, nrz, nsx, nsy, nsz);
}

World::~World() {
	//clean up
	for(unsigned int i = 0; i < _models.size(); i++) {
		delete _models[i];
	}
	
	//delete groundMotionState
	//delete body->getMotionState();
	//body is deleted in Physics
	//delete body;
	//delete trimesh;
	//delete shape;
}